/*--------------------------------------------------------------
    Authenticating IP Gateway Controller v1.4     authipgate.c
      Written  by H.Goto, May 2001       (formerly chkclogin.c)
      Modified by H.Goto, Aug 1, 2001
      Modified by H.Goto, Jun 4, 2002
          e-mail: hgot@rd.isc.tohoku.ac.jp
--------------------------------------------------------------*/

#define		UTMPFILE	"/var/run/utmp"
#define		IPCHAINS	"/sbin/ipchains"
#define		IPTABLES	"/sbin/iptables"

#define		MAXAPHOSTS	512

#define		STAYCOUNTS	4

#include	<stdio.h>
#include	<stdlib.h>
#include	<utmp.h>
#include	<string.h>
#include	<sys/types.h>
#include	<sys/stat.h>

#define		MAXCMDLEN	(UT_HOSTSIZE + 80)

typedef struct {
	int	f_used;
	int	stay;
	int	stat_prev, stat;
	char	hostID[UT_HOSTSIZE];
} ClientHost;

static ClientHost	hostlist[MAXAPHOSTS];

static int	sw_debug = 0;		/* show debug information */
static int	sw_nop = 0;		/* debug without any ipf operation */
static int	sw_iptables = 1;

static int	interval_diff = 2;
static int	interval_curr = 4;



/*--------------------------------------------------------------
    Initialize IP firewall
--------------------------------------------------------------*/

static int init_ipfw(){
	char	tmpcmd[MAXCMDLEN];
	/* Initialization of IP Firewall must be made before */
	/* this program is run. Following code just makes it */
	/* sure that the policy is all-deny. */
	if ( sw_iptables ){
		sprintf(tmpcmd,"%s -P FORWARD DROP >/dev/null 2>&1",IPTABLES);
	}
	else{	sprintf(tmpcmd,"%s -P forward DENY >/dev/null 2>&1",IPCHAINS);
	}
	if ( sw_nop == 0 ){
		if ( 0 != system(tmpcmd) )  return(-1);
	}
	return(0);
}




/*--------------------------------------------------------------
    Accept a specified host
--------------------------------------------------------------*/

static accept_host(char *host){
	char	tmpcmd[MAXCMDLEN];
	int	retcode = 0;
	if ( sw_iptables ){
		sprintf(tmpcmd,"%s -A FORWARD -j ACCEPT -d %s >/dev/null 2>&1", \
			IPTABLES,host);
	}
	else{	sprintf(tmpcmd,"%s -A forward -j ACCEPT -d %s >/dev/null 2>&1", \
			IPCHAINS,host);
	}
	if ( sw_debug )  printf("%s\n",tmpcmd);
	if ( sw_nop == 0 ){
		if ( 0 != system(tmpcmd) )  retcode = -1;
	}
	if ( sw_iptables ){
		sprintf(tmpcmd,"%s -A FORWARD -j ACCEPT -s %s >/dev/null 2>&1", \
			IPTABLES,host);
	}
	else{	sprintf(tmpcmd,"%s -A forward -j ACCEPT -s %s >/dev/null 2>&1", \
			IPCHAINS,host);
	}
	if ( sw_debug )  printf("%s\n",tmpcmd);
	if ( sw_nop == 0 ){
		if ( 0 != system(tmpcmd) )  retcode = -1;
	}
	return(retcode);
}




/*--------------------------------------------------------------
    Deny a specified host
--------------------------------------------------------------*/

static deny_host(char *host){
	char	tmpcmd[MAXCMDLEN];
	int	retcode = 0;
	if ( sw_iptables ){
		sprintf(tmpcmd,"%s -D FORWARD -j ACCEPT -s %s >/dev/null 2>&1", \
			IPTABLES,host);
	}
	else{	sprintf(tmpcmd,"%s -D forward -j ACCEPT -s %s >/dev/null 2>&1", \
			IPCHAINS,host);
	}
	if ( sw_debug )  printf("%s\n",tmpcmd);
	if ( sw_nop == 0 ){
		if ( 0 != system(tmpcmd) )  retcode = -1;
	}
	if ( sw_iptables ){
		sprintf(tmpcmd,"%s -D FORWARD -j ACCEPT -d %s >/dev/null 2>&1", \
			IPTABLES,host);
	}
	else{	sprintf(tmpcmd,"%s -D forward -j ACCEPT -d %s >/dev/null 2>&1", \
			IPCHAINS,host);
	}
	if ( sw_debug )  printf("%s\n",tmpcmd);
	if ( sw_nop == 0 ){
		if ( 0 != system(tmpcmd) )  retcode = -1;
	}
	return(retcode);
}




/*--------------------------------------------------------------
    Check login status
        If someone logins from a host and if the host has been
        registered before, this function returns the internal
        serial number of the registered host.
        Otherwise, this function registers the host and returns
        the serial number. If no more entry is available in the
        host table, this function returns -1.
--------------------------------------------------------------*/

static int chklogin(char *host){
	int	i,n;
	n = -1;
	for ( i=0 ; i<MAXAPHOSTS ; i++ ){
		if ( hostlist[i].f_used ){
			if ( 0 == strcmp(hostlist[i].hostID, host) )  return(i);
		}
		else{	if ( n == -1 )  n = i;
		}
	}
	if ( n == -1 )  return(-1);	/* table overflow */

	/* register new host */
	strncpy(hostlist[n].hostID, host, UT_HOSTSIZE);
	hostlist[n].hostID[UT_HOSTSIZE -1] = '\0';
	hostlist[n].f_used = 1;
	hostlist[n].stat_prev = 0;
	hostlist[n].stay = STAYCOUNTS;
	if ( sw_debug ){
		printf("Host '%s' has been added to the host table.\n",hostlist[n].hostID);
	}
	return(n);
}




/*--------------------------------------------------------------
    Main routine
--------------------------------------------------------------*/

int main(int ac,char *av[]){
	struct utmp	UT;
	int	i,hostid;
	FILE	*fp;
	int	repcount;
	int	f_startup = 1;
	struct stat	utmpstat0, utmpstat;

	if ( -1 == stat(IPTABLES,&utmpstat0) )
		sw_iptables = 0;	/* Use ipchains instead if iptables is not found. */

	for ( i=1 ; i<ac ; i++ ){
		if ( 0 == strcmp(av[i],"-n") ){ sw_nop = 1; continue; }
		if ( 0 == strcmp(av[i],"-d") ){ sw_debug = 1; continue; }
		if ( 0 == strcmp(av[i],"-chains") ){ sw_iptables = 0; continue; }
		if ( 0 == strcmp(av[i],"-tables") ){ sw_iptables = 1; continue; }
		if ( 0 == strcmp(av[i],"-h") || 0 == strcmp(av[i],"-help") ){
			fputs("Authenticating IP Gateway Controller v1.4 \n",stdout);
			fputs("Copyright (C) 2001, 2002 Hideaki Goto \n\n",stdout);
			fputs("usage: authipgate [-n] [-d] [-chains|-tables]\n",stdout);
			fputs("    -n : Does not issue ipchains/iptables commands. \n",stdout);
			fputs("    -d : Show debug information. \n",stdout);
			fputs("    -chains : Force to use ipchains. \n",stdout);
			fputs("    -tables : Force to use iptables. \n",stdout);
			return(0);
		}
		fputs("Undefined option.\n",stderr);
		exit(1);
	}

	if ( init_ipfw() ){
		fputs("Failed in initializing IP Firewall.\n",stderr);
		exit(1);
	}

	for ( i=0 ; i<MAXAPHOSTS ; i++ ){
		hostlist[i].f_used = 0;
		hostlist[i].stat_prev = hostlist[i].stat = 0;
	}

	stat(UTMPFILE,&utmpstat0);  utmpstat0.st_mtime--;

	if ( NULL == (fp = fopen(UTMPFILE,"r")) )  return(1);

	repcount = interval_curr;
	while(1){
		if ( ++repcount >= interval_curr )  repcount = 0;

		stat(UTMPFILE,&utmpstat);
		if ( utmpstat0.st_mtime == utmpstat.st_mtime && repcount != 0 ){
			if ( interval_diff > 0 )  sleep(interval_diff);
			continue;
		}
		utmpstat0 = utmpstat;

		for ( i=0 ; i<MAXAPHOSTS ; i++ )  hostlist[i].stat = 0;
		for ( i=0 ; i<2 ; i++ ){	/* read utmp file twice */
			fseek(fp,0,SEEK_SET);
			while( 1 == fread((void *)&UT,sizeof(struct utmp),1,fp) ){
				/* not a user process */
				if ( UT.ut_type != USER_PROCESS )  continue;

				/* not a remote session */
				if ( strlen(UT.ut_host) == 0 )  continue;

				if ( -1 == (hostid = chklogin(UT.ut_host) ) )  continue;
				hostlist[hostid].stat = 1;
//				printf("%4d %s %s %s\n",hostid,UT.ut_user,UT.ut_host,UT.ut_line);
			}
			if ( i == 0 )  sleep(1);
		}

		/* show current info. */
		if ( repcount == 0 ){
			for ( i=0 ; i<MAXAPHOSTS ; i++ ){
				if ( hostlist[i].f_used == 0 )  continue;
				if ( hostlist[i].stat == 0 ){
					if ( sw_debug ){
						printf("curr - %s\n",hostlist[i].hostID);
					}
					deny_host(hostlist[i].hostID);
				/* Note: Although multiple deletions of "accept" */
				/*       rule cause an error, we allow the       */
				/*       deletions in order to keep the system   */
				/*       secure.                                 */
					if ( --hostlist[i].stay <= 0 ){
						hostlist[i].f_used = 0;
						/* delete the entry */
						if ( sw_debug ){
							printf("Host '%s' has been removed from the host table.\n",hostlist[i].hostID);
						}
					}
				}
				else if ( hostlist[i].stat != 0 && f_startup ){
					if ( sw_debug ){
						printf("curr + %s\n",hostlist[i].hostID);
					}
					accept_host(hostlist[i].hostID);
				/* Note: Multiple additions of "accept" rule   */
				/*       must be avoided.  Only an addition    */
				/*       is done just after flushing all       */
				/*       policies at startup for duplicating   */
				/*       current login status.                 */
					hostlist[i].stat_prev = 1;
					hostlist[i].stay = STAYCOUNTS;
				}
			}
			f_startup = 0;
		}

		/* show differential info. */
		for ( i=0 ; i<MAXAPHOSTS ; i++ ){
			if ( hostlist[i].f_used == 0 )  continue;
			if ( hostlist[i].stat_prev == hostlist[i].stat )  continue;
			if ( hostlist[i].stat_prev == 0 ){
				if ( sw_debug ){
					printf("diff + %s\n",hostlist[i].hostID);
				}
				accept_host(hostlist[i].hostID);
				hostlist[i].stay = STAYCOUNTS;
			}
			else{	if ( sw_debug ){
					printf("diff - %s\n",hostlist[i].hostID);
				}
				deny_host(hostlist[i].hostID);
				if ( --hostlist[i].stay <= 0 ){
					hostlist[i].f_used = 0;
					/* delete the entry */
					if ( sw_debug ){
						printf("Host '%s' has been removed from the host table.\n",hostlist[i].hostID);
					}
				}
			}
			hostlist[i].stat_prev = hostlist[i].stat;
		}

		if ( interval_diff > 1 ){
			sleep(interval_diff -1);
		}
	}

	fclose(fp);
	return(0);
}


